@@ -137,6 +137,7 @@ func DefaultOptions(overrides ...func(*Options)) *Options {
 						},
 						Vulncheck:                 ModeVulncheckOff,
 						DiagnosticsDelay:          1 * time.Second,
+						DiagnosticsTrigger:        DiagnosticsOnEdit,
 						AnalysisProgressReporting: true,
 					},
 					InlayHintOptions: InlayHintOptions{},
@@ -468,6 +469,9 @@ type DiagnosticOptions struct {
 	// This option must be set to a valid duration string, for example `"250ms"`.
 	DiagnosticsDelay time.Duration `status:"advanced"`
 
+	// DiagnosticsTrigger controls when to run diagnostics.
+	DiagnosticsTrigger DiagnosticsTrigger `status:"experimental"`
+
 	// AnalysisProgressReporting controls whether gopls sends progress
 	// notifications when construction of its index of analysis facts is taking a
 	// long time. Cancelling these notifications will cancel the indexing task,
@@ -595,6 +599,9 @@ type InternalOptions struct {
 	// LiteralCompletions controls whether literal candidates such as
 	// "&someStruct{}" are offered. Tests disable this flag to simplify
 	// their expected values.
+	//
+	// TODO(rfindley): this is almost unnecessary now. Remove it along with the
+	// old marker tests.
 	LiteralCompletions bool
 
 	// VerboseWorkDoneProgress controls whether the LSP server should send
@@ -807,6 +814,17 @@ const (
 	// TODO: VulncheckRequire, VulncheckCallgraph
 )
 
+type DiagnosticsTrigger string
+
+const (
+	// Trigger diagnostics on file edit and save. (default)
+	DiagnosticsOnEdit DiagnosticsTrigger = "Edit"
+	// Trigger diagnostics only on file save. Events like initial workspace load
+	// or configuration change will still trigger diagnostics.
+	DiagnosticsOnSave DiagnosticsTrigger = "Save"
+	// TODO: support "Manual"?
+)
+
 type OptionResults []OptionResult
 
 type OptionResult struct {
@@ -1181,6 +1199,7 @@ func (o *Options) set(name string, value interface{}, seen map[string]struct{})
 		result.setBool(&o.VerboseWorkDoneProgress)
 
 	case "tempModfile":
+		result.softErrorf("gopls setting \"tempModfile\" is deprecated.\nPlease comment on https://go.dev/issue/63537 if this impacts your workflow.")
 		result.setBool(&o.TempModfile)
 
 	case "showBugReports":
@@ -1207,6 +1226,7 @@ func (o *Options) set(name string, value interface{}, seen map[string]struct{})
 		result.setBool(&o.NoSemanticNumber)
 
 	case "expandWorkspaceToModule":
+		result.softErrorf("gopls setting \"expandWorkspaceToModule\" is deprecated.\nPlease comment on https://go.dev/issue/63536 if this impacts your workflow.")
 		result.setBool(&o.ExpandWorkspaceToModule)
 
 	case "experimentalPostfixCompletions":
@@ -1239,6 +1259,14 @@ func (o *Options) set(name string, value interface{}, seen map[string]struct{})
 	case "diagnosticsDelay":
 		result.setDuration(&o.DiagnosticsDelay)
 
+	case "diagnosticsTrigger":
+		if s, ok := result.asOneOf(
+			string(DiagnosticsOnEdit),
+			string(DiagnosticsOnSave),
+		); ok {
+			o.DiagnosticsTrigger = DiagnosticsTrigger(s)
+		}
+
 	case "analysisProgressReporting":
 		result.setBool(&o.AnalysisProgressReporting)
 
@@ -1356,6 +1384,11 @@ func (r *OptionResult) deprecated(replacement string) {
 	r.Error = &SoftError{msg}
 }
 
+// softErrorf reports a soft error related to the current option.
+func (r *OptionResult) softErrorf(format string, args ...any) {
+	r.Error = &SoftError{fmt.Sprintf(format, args...)}
+}
+
 // unexpected reports that the current setting is not known to gopls.
 func (r *OptionResult) unexpected() {
 	r.Error = fmt.Errorf("unexpected gopls setting %q", r.Name)
@@ -1599,7 +1632,7 @@ func defaultAnalyzers() map[string]*Analyzer {
 		atomicalign.Analyzer.Name:      {Analyzer: atomicalign.Analyzer, Enabled: true},
 		deepequalerrors.Analyzer.Name:  {Analyzer: deepequalerrors.Analyzer, Enabled: true},
 		fieldalignment.Analyzer.Name:   {Analyzer: fieldalignment.Analyzer, Enabled: false},
-		nilness.Analyzer.Name:          {Analyzer: nilness.Analyzer, Enabled: false},
+		nilness.Analyzer.Name:          {Analyzer: nilness.Analyzer, Enabled: true},
 		shadow.Analyzer.Name:           {Analyzer: shadow.Analyzer, Enabled: false},
 		sortslice.Analyzer.Name:        {Analyzer: sortslice.Analyzer, Enabled: true},
 		testinggoroutine.Analyzer.Name: {Analyzer: testinggoroutine.Analyzer, Enabled: true},
@@ -1688,7 +1721,7 @@ var parBreakRE = regexp.MustCompile("\n{2,}")
 
 func collectEnums(opt *OptionJSON) string {
 	var b strings.Builder
-	write := func(name, doc string, index, len int) {
+	write := func(name, doc string) {
 		if doc != "" {
 			unbroken := parBreakRE.ReplaceAllString(doc, "\\\n")
 			fmt.Fprintf(&b, "* %s\n", strings.TrimSpace(unbroken))
@@ -1698,13 +1731,13 @@ func collectEnums(opt *OptionJSON) string {
 	}
 	if len(opt.EnumValues) > 0 && opt.Type == "enum" {
 		b.WriteString("\nMust be one of:\n\n")
-		for i, val := range opt.EnumValues {
-			write(val.Value, val.Doc, i, len(opt.EnumValues))
+		for _, val := range opt.EnumValues {
+			write(val.Value, val.Doc)
 		}
 	} else if len(opt.EnumKeys.Keys) > 0 && shouldShowEnumKeysInSettings(opt.Name) {
 		b.WriteString("\nCan contain any of:\n\n")
-		for i, val := range opt.EnumKeys.Keys {
-			write(val.Name, val.Doc, i, len(opt.EnumKeys.Keys))
+		for _, val := range opt.EnumKeys.Keys {
+			write(val.Name, val.Doc)
 		}
 	}
 	return b.String()
