@@ -284,24 +284,44 @@ func SnapshotLabels(snapshot Snapshot) []label.Label {
 	return []label.Label{tag.Snapshot.Of(snapshot.SequenceID()), tag.Directory.Of(snapshot.View().Folder())}
 }
 
-// NarrowestPackageForFile is a convenience function that selects the
-// narrowest non-ITV package to which this file belongs, type-checks
-// it in the requested mode (full or workspace), and returns it, along
-// with the parse tree of that file.
+// NarrowestPackageForFile is a convenience function that selects the narrowest
+// non-ITV package to which this file belongs, type-checks it in the requested
+// mode (full or workspace), and returns it, along with the parse tree of that
+// file.
 //
-// The "narrowest" package is the one with the fewest number of files
-// that includes the given file. This solves the problem of test
-// variants, as the test will have more files than the non-test package.
-// (Historically the preference was a parameter but widest was almost
-// never needed.)
+// The "narrowest" package is the one with the fewest number of files that
+// includes the given file. This solves the problem of test variants, as the
+// test will have more files than the non-test package.
 //
-// An intermediate test variant (ITV) package has identical source
-// to a regular package but resolves imports differently.
-// gopls should never need to type-check them.
+// An intermediate test variant (ITV) package has identical source to a regular
+// package but resolves imports differently. gopls should never need to
+// type-check them.
 //
-// Type-checking is expensive. Call snapshot.ParseGo if all you need
-// is a parse tree, or snapshot.MetadataForFile if you only need metadata.
+// Type-checking is expensive. Call snapshot.ParseGo if all you need is a parse
+// tree, or snapshot.MetadataForFile if you only need metadata.
 func NarrowestPackageForFile(ctx context.Context, snapshot Snapshot, uri span.URI) (Package, *ParsedGoFile, error) {
+	return selectPackageForFile(ctx, snapshot, uri, func(metas []*Metadata) *Metadata { return metas[0] })
+}
+
+// WidestPackageForFile is a convenience function that selects the widest
+// non-ITV package to which this file belongs, type-checks it in the requested
+// mode (full or workspace), and returns it, along with the parse tree of that
+// file.
+//
+// The "widest" package is the one with the most number of files that includes
+// the given file. Which is the test variant if one exists.
+//
+// An intermediate test variant (ITV) package has identical source to a regular
+// package but resolves imports differently. gopls should never need to
+// type-check them.
+//
+// Type-checking is expensive. Call snapshot.ParseGo if all you need is a parse
+// tree, or snapshot.MetadataForFile if you only need metadata.
+func WidestPackageForFile(ctx context.Context, snapshot Snapshot, uri span.URI) (Package, *ParsedGoFile, error) {
+	return selectPackageForFile(ctx, snapshot, uri, func(metas []*Metadata) *Metadata { return metas[len(metas)-1] })
+}
+
+func selectPackageForFile(ctx context.Context, snapshot Snapshot, uri span.URI, selector func([]*Metadata) *Metadata) (Package, *ParsedGoFile, error) {
 	metas, err := snapshot.MetadataForFile(ctx, uri)
 	if err != nil {
 		return nil, nil, err
@@ -310,8 +330,8 @@ func NarrowestPackageForFile(ctx context.Context, snapshot Snapshot, uri span.UR
 	if len(metas) == 0 {
 		return nil, nil, fmt.Errorf("no package metadata for file %s", uri)
 	}
-	narrowest := metas[0]
-	pkgs, err := snapshot.TypeCheck(ctx, narrowest.ID)
+	md := selector(metas)
+	pkgs, err := snapshot.TypeCheck(ctx, md.ID)
 	if err != nil {
 		return nil, nil, err
 	}
@@ -352,9 +372,16 @@ func (m InvocationFlags) AllowNetwork() bool {
 	return m&AllowNetwork != 0
 }
 
-// View represents a single workspace.
-// This is the level at which we maintain configuration like working directory
-// and build tags.
+// View represents a single build context for a workspace.
+//
+// A unique build is determined by the workspace folder along with a Go
+// environment (GOOS, GOARCH, GOWORK, etc).
+//
+// Additionally, the View holds a pointer to the current state of that build
+// (the Snapshot).
+//
+// TODO(rfindley): move all other state such as module upgrades into the
+// Snapshot.
 type View interface {
 	// ID returns a globally unique identifier for this view.
 	ID() string
@@ -711,7 +738,6 @@ const (
 	Save
 	Create
 	Delete
-	InvalidateMetadata
 )
 
 func (a FileAction) String() string {
@@ -728,8 +754,6 @@ func (a FileAction) String() string {
 		return "Create"
 	case Delete:
 		return "Delete"
-	case InvalidateMetadata:
-		return "InvalidateMetadata"
 	default:
 		return "Unknown"
 	}
